WhoHasConfig = {
   enabled   = 1,
   totals    = 1,
   stacks    = 1,
   inbox     = 1,
   keyring   = 1,
   bags      = 1,
   equipment = 1
}

WhoHas = {}

WhoHas.state = {
   savedName = "";
   player = "";
   realm = "";
   tooltipText = {};
   altCache = {};
   playerCache = {};
   inventoryChanged = 1;
}

-- these are internal strings, not for display
WhoHas.categories = {
   "Inventory",
   "Bank",
   "Inbox",
   "Keyring",
   "Equipment",
   "InvBags",
   "BankBags",
}

-------------------------------------------------------------------------------
-- OnLoad
-------------------------------------------------------------------------------

function WhoHas.OnLoad()
   SlashCmdList["WHOHAS"] = WhoHas.ShowConfigFrame;
   tinsert(UISpecialFrames, "WhoHasConfigFrame");

   if (PossessionsData) then
      WhoHas.ScanAlts = WhoHas.ScanAltsPoss;
      WhoHas.ScanPlayer = WhoHas.ScanPlayerPoss;
   elseif (myProfile) then
      WhoHas.ScanAlts = WhoHas.ScanAltsCP;
      WhoHas.ScanPlayer = WhoHas.ScanPlayerCP;
   else
      WhoHas.ScanAlts = WhoHas.DoNothing;
      WhoHas.ScanPlayer = WhoHas.DoNothing;
   end

   WhoHas.Orig_SetItemRef = SetItemRef
   SetItemRef             = WhoHas.SetItemRef

   if (Possessions_ItemButton_OnEnter) then
      WhoHas_Possessions_ItemButton_OnEnter = Possessions_ItemButton_OnEnter
      Possessions_ItemButton_OnEnter        = WhoHas.Possessions_ItemButton_OnEnter
   end

   for event in pairs(WhoHas.Events) do
      this:RegisterEvent(event);
   end
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

WhoHas.Events = {}

function WhoHas.OnEvent()
   local func = WhoHas.Events[event];
   if (func) then
      func(arg1, arg2);
   end
end

function WhoHas.Events.ADDON_LOADED(name)
   if (name == "WhoHas") then
      WhoHas.state.player = UnitName("player");
      WhoHas.state.realm = GetRealmName();
      WhoHas.ScanAlts();
   end
end

function WhoHas.InventoryChanged()
   WhoHas.state.inventoryChanged = 1;
end

WhoHas.Events.UNIT_INVENTORY_CHANGED = WhoHas.InventoryChanged;
WhoHas.Events.BAG_UPDATE             = WhoHas.InventoryChanged;

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function WhoHas.OnShow()
   WhoHas.ShowTooltip(GameTooltip);
end

function WhoHas.SetItemRef(args)
   WhoHas.Orig_SetItemRef(args)
   WhoHas.ShowTooltip(ItemRefTooltip);
end

function WhoHas.Possessions_ItemButton_OnEnter(args)
   -- don't doctor tooltips inside of Possessions
   WhoHas.skip = true;
   WhoHas_Possessions_ItemButton_OnEnter(args)
   WhoHas.skip = nil;
end

function WhoHas.ShowConfigFrame()
   WhoHasConfigFrame:Show()
end

function WhoHas.ShowTooltip(tooltip, link)
   if (tooltip and WhoHasConfig.enabled and not WhoHas.skip) then
      local name = getglobal(tooltip:GetName().."TextLeft1"):GetText();
      if (not name or name == "") then
         return;
      end

      if (WhoHas.state.inventoryChanged) then
         WhoHas.ScanPlayer();
         WhoHas.state.inventoryChanged = nil;
         WhoHas.state.savedName = "";
      end

      if (name ~= WhoHas.state.savedName) then
         WhoHas.state.tooltipText = {};
         WhoHas.state.savedName = name;
         WhoHas.GetText(name, WhoHas.state.tooltipText);
      end

      for i, line in ipairs(WhoHas.state.tooltipText) do
         tooltip:AddLine(line);
      end
      tooltip:Show();
   end
end

function WhoHas.GetText(name, text)
   local total = WhoHas.ListOwners(name, text);
   if (WhoHasConfig.totals and total > 0) then
      table.insert(text, string.format(WhoHas.formats.total, total));
   end
   if (WhoHasConfig.stacks) then
      local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, invTexture = GetItemInfo(name);
      if (itemStackCount and itemStackCount > 1) then
         table.insert(text, string.format(WhoHas.formats.stack, itemStackCount));
      end
   end
end

function WhoHas.ListOwners(name, text)
   local total = 0;
   if (WhoHas.state.playerCache[name]) then
      total = total + WhoHas.ListChar(WhoHas.state.player, WhoHas.state.playerCache[name], text);
   end
   for charName, charData in pairs(WhoHas.state.altCache) do
      if (charData[name]) then
         total = total + WhoHas.ListChar(charName, charData[name], text);
      end
   end
   return total;
end

function WhoHas.ListChar(charName, charData, text)
   local total = 0;
   for i, category in ipairs(WhoHas.categories) do
      local count = charData[category];
      if count and count > 0 then
         table.insert(text, string.format(WhoHas.formats[category], count, charName));
         total = total + count;
      end
   end
   return total;
end

WhoHas.Poss = {}
WhoHas.Poss.Lauchsuppe = {
   Inventory = { 0, 1, 2, 3, 4 },
   Bank      = { -1, 5, 6, 7, 8, 9, 10, 11 },
   Keyring   = { -2 },
   Equipment = { -3 },
   Inbox     = { -4 },
   InvBags   = { -5 },
   BankBags  = { -6 }
}

WhoHas.Poss.Siz = {
   Inventory = { 0, 1, 2, 3, 4 },
   Bank      = { -1, 5, 6, 7, 8, 9, 10, 11 },
   Equipment = { -2 },
   Inbox     = { -3 },
   Keyring   = { -4 }
}

function WhoHas.ScanPlayerPoss()
   local slots;

   if (POSS_USEDBANKBAGS_CONTAINER) then
      slots = WhoHas.Poss.Lauchsuppe;
   else
      slots = WhoHas.Poss.Siz;
   end

   local charData;
   if (PossessionsData[WhoHas.state.realm]) then
      charData = PossessionsData[WhoHas.state.realm][string.lower(WhoHas.state.player)];
   end
   if (charData) then
      WhoHas.state.playerCache = {};
      WhoHas.ScanCharPoss(WhoHas.state.player, charData, slots, WhoHas.state.playerCache);
   end
end

function WhoHas.ScanAltsPoss()
   local slots;

   if (POSS_USEDBANKBAGS_CONTAINER) then
      slots = WhoHas.Poss.Lauchsuppe;
   else
      slots = WhoHas.Poss.Siz;
   end

   if (PossessionsData[WhoHas.state.realm]) then
      for charName, charData in pairs(PossessionsData[WhoHas.state.realm]) do
         -- Possessions lower-cases character names, annoyingly
         charName = string.upper(string.sub(charName, 1, 1)) .. string.sub(charName, 2);
         if (charName ~= WhoHas.state.player) then
            WhoHas.state.altCache[charName] = {};
            WhoHas.ScanCharPoss(charName, charData, slots, WhoHas.state.altCache[charName]);
         end
      end
   end
end

function WhoHas.ScanCharPoss(charName, charData, slots, cache)
   WhoHas.ScanBagsPoss(charName, "Inventory", charData.items, slots.Inventory, cache);
   WhoHas.ScanBagsPoss(charName, "Bank", charData.items, slots.Bank, cache);
   if (WhoHasConfig.inbox) then
      WhoHas.ScanBagsPoss(charName, "Inbox", charData.items, slots.Inbox, cache);
   end
   if (WhoHasConfig.keyring) then
      WhoHas.ScanBagsPoss(charName, "Keyring", charData.items, slots.Keyring, cache);
   end
   if (WhoHasConfig.equipment) then
      WhoHas.ScanBagsPoss(charName, "Equipment", charData.items, slots.Equipment, cache);
   end
   if (WhoHasConfig.bags and slots.InvBags) then
      WhoHas.ScanBagsPoss(charName, "InvBags", charData.items, slots.InvBags, cache);
   end
   if (WhoHasConfig.bags and slots.BankBags) then
      WhoHas.ScanBagsPoss(charName, "BankBags", charData.items, slots.BankBags, cache);
   end
end

function WhoHas.ScanBagsPoss(char, slot, bags, bagIndex, cache)
   for _, index in pairs(bagIndex) do
      if (bags[index]) then
         for i, item in pairs(bags[index]) do
            if (item and item[1]) then
               local name = item[1];
               local count = item[3] or 1;
               if (not cache[name]) then
                  cache[name] = {};
               end
               cache[name][slot] = count + (cache[name][slot] or 0);
            end
         end
      end
   end
end

function WhoHas.ScanPlayerCP()
   local charData;
   if (myProfile[WhoHas.state.realm]) then
      charData = myProfile[WhoHas.state.realm].Character[WhoHas.state.player];
   end
   if (charData) then
      WhoHas.state.playerCache = {};
      WhoHas.doBagsCP(charName, charData.Inventory, "Inventory", WhoHas.formats.Inventory, WhoHas.state.playerCache);
      WhoHas.doBagsCP(charName, charData.Bank, "Bank", WhoHas.formats.Bank, WhoHas.state.playerCache);
      WhoHas.doInboxCP(charName, charData.MailBox, "Inbox", WhoHas.formats.Inbox, WhoHas.state.playerCache);
   end
end

function WhoHas.ScanAltsCP()
   for charName, charData in pairs(myProfile[GetRealmName()].Character) do
      if (charName ~= WhoHas.state.player) then
         WhoHas.state.altCache[charName] = {};
         WhoHas.doBagsCP(charName, charData.Inventory, "Inventory", WhoHas.formats.Inventory, WhoHas.state.altCache[charName]);
         WhoHas.doBagsCP(charName, charData.Bank, "Bank", WhoHas.formats.Bank, WhoHas.state.altCache[charName]);
         WhoHas.doInboxCP(charName, charData.MailBox, "Inbox", WhoHas.formats.Inbox, WhoHas.state.altCache[charName]);
      end
   end
end

function WhoHas.doBagsCP(char, bags, slot, format, cache)
   for bag, bagData in pairs(bags) do
      for i = 1, bagData.Slots do
         local item = bagData.Contents[i]
         if (item and item.Name) then
            local count = item.Quantity or 1;
            if (not cache[item.Name]) then
               cache[item.Name] = {};
            end
            cache[item.Name][slot] = count + (cache[item.Name][slot] or 0);
         end
      end
   end
end

function WhoHas.doInboxCP(char, inbox, slot, format, cache)
   for i, item in ipairs(inbox) do
      if (item and item.Name) then
         local count = item.Quantity or 1;
         if (not cache[item.Name]) then
            cache[item.Name] = {};
         end
         cache[item.Name][slot] = count + (cache[item.Name][slot] or 0);
      end
   end
end

function WhoHas.DoNothing()
end
